package Maze::AStar;
use strict;
my $MAXSTEPS = 1000;

sub new {
  my($class, $map) = @_;
  my $self = bless { map => $map }, $class;
  return $self;
}

sub findpath {
  my($self, $source, $destination) = @_;
  my $map = $self->{map};

  my($sx, $sy) = ($source->[0], $source->[1]);
  my($dx, $dy) = ($destination->[0], $destination->[1]);

#print "Source: $sx, $sy\n";
#print "  Dest: $dx, $dy\n";

  my $edge = [[$sx, $sy]]; # list of edges to consider
  my $done; # whether we've done this cell before
  my $dist; # distance from the start to a cell
  my $heuristic; # A* heuristic

  $dist->[$sx]->[$sy] = 0;

  my $step = 0;

  my %weight;

  while (1) {
#    warn "\nStep $step\n";


    $edge = [sort { $heuristic->[$a->[0]]->[$a->[1]] <=> $heuristic->[$b->[0]]->[$b->[1]] } @$edge];

    my $c = shift @$edge;

    my($cx, $cy) = ($c->[0], $c->[1]);
    next if $done->[$cx]->[$cy]++;
    next if $map->[$cx]->[$cy] =~ /[\#~]/;

    my $distsofar = $dist->[$cx]->[$cy];
    #     warn "C: $cx, $cy ($distsofar)\n";

    my @next = (
      [$cx-1,$cy],
      [$cx+1,$cy],
      [$cx,$cy-1],
      [$cx,$cy+1],
    );


    foreach my $now (@next) {
      my($nx, $ny) = ($now->[0], $now->[1]);

      next if $map->[$nx]->[$ny] =~ /[\#~]/;
      next if defined $dist->[$nx]->[$ny];

      #	warn "N: $nx, $ny\n";
      my $curdist = $dist->[$nx]->[$ny];
      my $newdist = $distsofar + 1;
      $dist->[$nx]->[$ny] = $newdist;
      my $hvalue = $newdist + (abs($dx - $nx) + abs($dy - $ny));
      $heuristic->[$nx]->[$ny] = $hvalue;

      if ($nx == $dx && $ny == $dy) {
	#	  warn "woohoo, found\n";
	my @path;

	while (1) {
	  push @path, [$nx, $ny];
	  last if $nx == $sx && $ny == $sy;
	  my @next = (
            [$nx-1,$ny],
            [$nx+1,$ny],
            [$nx,$ny-1],
            [$nx,$ny+1],
          );
	  my $lowest = 999;
	  foreach my $next (@next) {
	    my($nextx, $nexty) = ($next->[0], $next->[1]);
	    my $dist = $dist->[$nextx]->[$nexty];
	    next unless defined $dist;
	    #	      print "dist for $nextx, $nexty is $dist\n";
	    if ($dist < $lowest) {
	      $lowest = $dist;
	      ($nx, $ny) = ($nextx, $nexty);
	    }
	  }
	  #	    print "decided upon $nx, $ny\n";
	}
	pop @path;
	return [reverse @path];
      }
	
      if ($map->[$nx]->[$ny] !~ /[\#~]/ && !$done->[$nx]->[$ny]) {
	#          warn "Adding $nx, $ny\n";
	push @$edge, $now;
      }
    }

    if (!@$edge) {
      warn "there's no path, boo\n";
      return [];
    }

    # Keep on trying really hard
#    last if $step > $MAXSTEPS; # don't take too long

    }
}

