package Maze::Client;
use strict;
use IO::Socket::INET;

my $DEBUG = 1;

sub new {
  my($class, $server, $port) = @_;

  my $s = IO::Socket::INET->new(
    PeerAddr => $server,
    PeerPort => $port,
  ) || die "error connecting to server";

  my $self = {
    server => $s,
  };

  $self->{robots} = {};

  bless $self, $class;
  return $self;
}

sub readline {
  my $self = shift;
  my $text = $self->{server}->getline;
  chomp $text;
  warn "[READ: $text]\n" if $DEBUG;
  die "I died, boo" if $text =~ /died/;
  return $text;
}

sub writeline {
  my($self, $text) = @_;
  warn "[WRITE: $text]\n" if $DEBUG;
  $self->{server}->print("$text\n");
}

sub init {
  my $self = shift;
  my $data;

  $self->writeline("Player");

  my($width, $height) = split ' ', $self->readline;
  $data->{dimensions} = {
    width => $width,
    height => $height,
  };

  my $maze = "";
  foreach (1..$height) {
    $maze = $self->readline . "\n" . $maze;
  }
  $data->{maze} = $maze;

  my($id, $capacity, $fuel) = split ' ', $self->readline;
  $data->{conf} = {
    id => $id,
    capacity => $capacity,
    fuel => $fuel,
  };

  my @initial = split ' ', $self->readline;
  while (@initial) {
    my $id = shift(@initial);
    $id =~ s/^#//;		
    die "not X" unless shift(@initial) eq 'X';
    my $x = shift(@initial);
    die "not Y" unless shift(@initial) eq 'Y';
    my $y = shift(@initial);
    $self->{robots}->{$id} = 1;
    $data->{robots}->{$id} = {
      x => $x,
      y => $y,
    };
  }

  my $packages = $self->readline;
  if ($packages ne "") {
	my @pack = split (/\s/, $packages);
	my @each;
 	for(@pack) {
	    my $id = shift @pack;
	    my $packagex = shift @pack;
	    my $packagey = shift @pack;
	    my $weight = shift @pack;
	    push @each, [$packagex, $packagey, $weight, $id];
	}
	$data->{packages} = \@each;
  }

  return $data;
}

sub do {
  my($self, $decision) = @_;
  my $data;

  my $robots;

  # make a decision
  $self->writeline($decision);

  my $response = $self->readline;
  return undef unless $response; # no data?

  my @response = split '#', $response;
  shift @response; # knock out first empty one
  foreach my $response (@response) {
    next unless $response; # empty, nothing happened
    my @parts = split ' ', $response;
    my $id = shift @parts;
    $robots->{$id}++; # saw this robot this time aroud

    while (@parts) {
      my $command = shift @parts;
      next unless $command; # empty, nothing happened
      if ($command eq 'X') {
	# robot appears
	my $x = shift @parts;
	die "not Y" unless shift @parts eq "Y";
	my $y = shift @parts;
	push @{$data->{robots}->{$id}}, {
          action => 'appear',
          x => $x,
          y => $y,
        };
      } elsif ($command =~ /^[NESW]$/) {
	push @{$data->{robots}->{$id}}, {
          action => 'move',
          direction => $command,
        };
      } elsif ($command eq 'P') {
	my $pid = shift @parts;
	push @{$data->{robots}->{$id}}, {
          action => 'pick',
          package => $pid,
        };
      } elsif ($command eq 'D') {
	my $pid = shift @parts;
	push @{$data->{robots}->{$id}}, {
          action => 'drop',
          package => $pid,
        };
      } else {
	die "unknown command [$command]\n";
      }
    }
  }

  foreach my $rid (keys %{$self->{robots}}) {
    next if $robots->{$rid};
    # uh-oh, a robot has disappeared since last time
    push @{$data->{robots}->{$rid}}, {
      action => 'gone',
    };
  }
  $self->{robots} = $robots;

  # <strike>TODO!</strike>  Done!
  # Deal with packages.  Populate $data->{packages} with an arrayref to each.
  # Arrayref format:
  #   [ x, y, weight, id ]

  my $packages = $self->readline;
  if ($packages ne "") {
	my @pack = split (/\s/, $packages);
	my @each;
 	for(@pack) {
	    my $id = shift @pack;
	    my $packagex = shift @pack;
	    my $packagey = shift @pack;
	    my $weight = shift @pack;
	    push @each, [$packagex, $packagey, $weight, $id];
	}
	$data->{packages} = \@each;
  }
  return $data;
}

1;
