#!/usr/bin/perl -w

# "Sam sat on the ground and put his head in his hands.  'I wish I had never
# come here, and I don't want to see no more magic,' he said, and fell silent."

use strict;
use lib 'source';
use Maze::AStar;
use Maze::Client;
use Maze::Map;

my $SERVER = shift || "192.168.0.3";
my $PORT = shift || 9000;

my $client = Maze::Client->new($SERVER, $PORT);
my $data = $client->init;

my($width, $height) = (
  $data->{dimensions}->{width},
  $data->{dimensions}->{height},
);
my $maze = $data->{maze};
my $map = Maze::Map->new($width, $height, $maze);
my($id, $capacity, $fuel) = (
  $data->{conf}->{id},
  $data->{conf}->{capacity},
  $data->{conf}->{fuel},
);

warn "Dimensions: ($width, $height)\n";
warn "Maze:\n$maze";
warn "ID: $id, Capacity: $capacity, Fuel: $fuel\n";

my $robots = {};

foreach my $rid (keys %{$data->{robots}}) {
  my $r = $data->{robots}->{$rid};
  my($x, $y) = (
    $r->{x},
    $r->{y},
  );
  $robots->{$rid} = [$y, $x];
  warn "Robot $rid is at ($x, $y)\n";
}

my $astar = Maze::AStar->new($map, $width, $height);
my $packages = {};
my $wantpackages = {};
my $path = [];
my $score=0;
my($wanttox, $wanttoy, $wanttomove);
while (1) {

  my($mx, $my) = ($robots->{$id}->[0], $robots->{$id}->[1]);

  my $decision;

  my $dir;

  if (!@$path) {
    warn "path? @$path\n";
#    unless (keys %$packages) {


      if ($data->{packages}) {

	if (keys %$packages) {
	  warn "carrying packages, and just rolled over some more\n";
	}

        # This is where we decide what packages to pick up.
        # It looks like we should be choosing based on two things:
        #  1.  Where the packages are going to:
        #      (This minimises travel distance.  We should first sort
        #       the packages into the various destinations, then take
        #       as many as possible to the same destination.)
        # 2.   The weight of the packages:
        #      (This is a packing problem cf. travelling salesman.  We
        #       need to take packages that move us closest to our max
        #       capacity.)
	# [um, pick a random package for now]
	my $npackages = @{$data->{packages}};

	my @pickup; # pick packages to pick up

        # It looks like we should be choosing based on two things:
        #  1.  Where the packages are going to:

	my $currentweight = 0;
	foreach my $pid (keys %$packages) {
	  my $pdest = $packages->{$pid};
	  $currentweight += $pdest->[2];
	}
	warn "Current weight: $currentweight\n";
	my $extracapacity = $capacity - $currentweight;
	warn "Extra capacity: $extracapacity\n";

	my $p;
	foreach (@{$data->{packages}}) {
	    my ($px, $py, $pweight, $pid) = @{$_};
	    next if $pweight > $extracapacity;
	    push @{$p->{$px}->{$py}}, $_;
	}
	
	foreach my $pa (keys %$p) {
	    foreach my $pb (keys %{$p->{$pa}}) {
		push @pickup, sort
                  { $a->[2] <=> $b->[2] }
                  @{$p->{$pa}->{$pb}};
	    }
	}

	if (@pickup) {
	  warn "decide to pick up packages\n";
	  $decision = "1 Pick ";
	  foreach (@pickup) {
	    my ($px, $py, $pweight, $pid) = @{$_};
	    $wantpackages->{$pid} = [ $py, $px, $pweight ];
	    $decision .= "$pid ";
	  }
	  chop $decision;
	}
      }
#    }

  }

  if (!$decision) {
    warn "no decision atm\n";
    if (!@$path) {
      warn "no path atm\n";
      if (keys %$packages) {
	warn "have packages atm\n";
	my @todrop;
	# see if any packages are supposed to be dropped here
	foreach my $pid (keys %$packages) {
	  my $pdest = $packages->{$pid};
	  if ($pdest->[0] == $mx and $pdest->[1] == $my) {
	    push @todrop, $pid;
	  }
	}
	if (@todrop) {
	  warn "decide to drop @todrop\n";
      my ($bid, $sum); $bid=0; $sum=0;
      foreach (@todrop) { $sum += $packages->{$_}->[2]; }
      warn "Dropping $sum worth of packages.\n";
      $bid = robotaround($robots, $mx, $my) ? $sum : 1;
      $decision = "$bid Drop " . join(", ", @todrop);
	} else {
	  # pick a package to walk towards
	  my $pdest = (values %$packages)[0];
	  my($px, $py) = ($pdest->[0], $pdest->[1]);
	  warn "astaring to $px, $py\n";
	  warn "path to drop packages\n";
	  $path = $astar->findpath([$mx, $my], $pdest);
	  unless (@$path) {
	    my @choices = ("1 Move N", "1 Move S", "1 Move E", "1 Move W");
	    $decision = $choices[int(rand(@choices))];
	    warn "No idea what to do: randomly picking $decision\n";
	  }
	}
      } else {
	warn "have no packages atm\n";
	my @homebases = find_homebases($height, $width, $map);
	my $dest = $homebases[int(rand(@homebases))];
	my($dx, $dy) = ($dest->[0], $dest->[1]);
	warn "astaring to $dx, $dy\n";
	warn "path to pick up packages\n";
        if ($mx != $dx || $my != $dy) {
	  $path = $astar->findpath([$mx, $my], $dest);
	  unless (@$path) {
	    my @choices = ("1 Move N", "1 Move S", "1 Move E", "1 Move W");
	    $decision = $choices[int(rand(@choices))];
	    warn "No idea what to do: randomly picking $decision\n";
	  }
	}
      }
    }
  }

 HERE:

  if ((!$decision) && @$path) {
    my $next = shift @$path;
    my($nx, $ny) = ($next->[0], $next->[1]);
    ($wanttox, $wanttoy, $wanttomove) = ($nx, $ny, 1);
    my $bid = 1;
    if (robotaround($robots, $mx, $my)) {
      warn "Uhoh, there's a robot around\n";
      if (wateraround($map, $mx, $my)) {
	warn "Uhoh, there's also water around, bid high\n";
	$bid = 50;
      }
    }
    if ($nx > $mx) {
      $decision = "$bid Move N";
    } elsif ($nx < $mx) {
      $decision = "$bid Move S";
    } elsif ($ny > $my) {
      $decision = "$bid Move E";
    } elsif ($ny < $my) {
      $decision = "$bid Move W";
    } else {
      warn "uhoh, no dir";
	my @homebases = find_homebases($height, $width, $map);
	my $dest = $homebases[int(rand(@homebases))];
	my($dx, $dy) = ($dest->[0], $dest->[1]);
	warn "astaring to $dx, $dy\n";
	warn "path to pick up packages\n";
        if ($mx != $dx || $my != $dy) {
	  $path = $astar->findpath([$mx, $my], $dest);
	  unless (@$path) {
	    my @choices = ("1 Move N", "1 Move S", "1 Move E", "1 Move W");
	    $decision = $choices[int(rand(@choices))];
	    warn "No idea what to do: randomly picking $decision\n";
	  }
	  goto HERE;
	}
    }
  }

  unless ($decision) {
    my @choices = ("1 Move N", "1 Move S", "1 Move E", "1 Move W");
    $decision = $choices[int(rand(@choices))];
    warn "No idea what to do: randomly picking $decision\n";
  }

  warn "Decision: $decision\n";
  $data = $client->do($decision);
  $fuel--;
#  if (not defined $data) {
#    die "End of game!";
#  }

  foreach my $rid (keys %{$data->{robots}}) {
    my @r = @{$data->{robots}->{$rid}};
    foreach my $r (@r) {
      my $action = $r->{action};
      if ($action eq 'appear') {
        my($x, $y) = ($r->{x}, $r->{y});
	warn "Robot $rid appeared at $x, $y\n";
	$robots->{$rid} = [$y, $x];
      } elsif ($action eq 'gone') {
	warn "Robot $rid disappeared\n";
	delete $robots->{$rid};
      } elsif ($action eq "move") {
	my $direction = $r->{direction};
	warn "Robot $rid moved $direction\n";
	domove($direction, $robots->{$rid});
	if ($rid == $id) {
	  my($rx, $ry) = ($robots->{$rid}->[0], $robots->{$rid}->[1]);
	  if ($wanttomove && ($wanttox != $rx || $wanttoy != $ry)) {
	    warn "! Bumped off target from ($wanttox, $wanttoy) to ($rx, $ry)!\n";
	    $path = [];
	    $wanttomove = 0;
	  }
	}
      } elsif ($action eq "pick") {
	my $package = $r->{package};
	warn "Robot $rid picked up package $package\n";
	if ($rid == $id) {
	  warn "I've picked up $package!\n";
	  $packages->{$package} = $wantpackages->{$package};
	}
      } elsif ($action eq "drop") {
	my $package = $r->{package};
	if ($rid == $id) {
 	  warn "rid = id: $rid, $id\n";
          my $p = $packages->{$package};
	  my $px = $p->[0]; my $py = $p->[1];
 	  if (($px == $mx) and ($py == $my)) {
            $score += $p->[2];
          }
	delete $packages->{$package};
        }
	warn "Robot $rid dropped $package\n";
      } else {
	warn "Don't know about action $action!?\n";
      }
    }
  }

}

sub domove {
  my($direction, $r) = @_;
  if ($direction eq 'N') {
    $r->[0]++;
  } elsif ($direction eq 'S') {
    $r->[0]--;
  } elsif ($direction eq 'E') {
    $r->[1]++;
  } elsif ($direction eq 'W') {
    $r->[1]--;
  }
  return $r;
}

sub find_homebases {
    my ($height, $width, $map) = @_;
    my @homebases;
    my ($xs, $ys, $xcount, $ycount);
    $ycount = $height+1;
    foreach $ys (reverse @$map) {
	$xcount = 0;
	foreach $xs (@$ys) {
	    if ($xs eq '@') { 
		push @homebases, [$ycount, $xcount];
	    }
	    $xcount++;
	}
	$ycount--;
    }
    return @homebases;
}

sub robotaround {
    my ($robots, $mx, $my) = @_;
    my @next = (
	[$mx-1,$my],
	[$mx+1,$my],
	[$mx,$my-1],
	[$mx,$my+1]
    );

    foreach (@next) {
	my ($mx, $my) = ($_->[0], $_->[1]);
        foreach my $id (keys %$robots) {

            my($rx, $ry) = ($robots->{$id}->[0], $robots->{$id}->[1]);
            return 1 if $rx == $mx && $ry == $my;
        }
    }
    return 0;
}

sub wateraround {
    my ($map, $mx, $my) = @_;
    my @next = (
	[$mx-1,$my],
	[$mx+1,$my],
	[$mx,$my-1],
	[$mx,$my+1]
    );

    foreach (@next) {
	my ($mx, $my) = ($_->[0], $_->[1]);
	return 1 if $map->[$mx]->[$my] eq '~';
    }
    return 0;
}
